//
// mhuffyuv v1.3.1, by Ben Rudiak-Gould.
// http://www.math.berkeley.edu/~benrg/mhuffyuv.html
//
// Based on MSYUV sample code, which is:
// Copyright (c) 1993 Microsoft Corporation.
// All Rights Reserved.
//
// Changes copyright 2000 Ben Rudiak-Gould, and distributed under
// the terms of the GNU General Public License, v2 or later.  See
// http://www.gnu.org/copyleft/gpl.html.
//
// I edit these files in 10-point Verdana, a proportionally-spaced font.
// You may notice formatting oddities if you use a monospaced font.
//


#include "mhuffyuv.h"
#include "resource.h"


TCHAR szDescription[] = TEXT("mhuffyuv v1.3.1");
TCHAR szName[]        = TEXT("mhuffyuv");




const int cyb = int(0.114*219/255*65536+0.5);
const int cyg = int(0.587*219/255*65536+0.5);
const int cyr = int(0.299*219/255*65536+0.5);


#define VERSION         0x00010003      // 1.3


/********************************************************************
********************************************************************/
CodecInst::CodecInst() 
{ 
  yuy2_buffer[0]=NULL;
  yuy2_buffer[1]=NULL;
  icinfo_old=NULL;
  icinfo_old2=NULL;
  icinfo_sel=3;
  FrameCount=0;
}//CodecInst

/********************************************************************
********************************************************************/

void Msg(const char fmt[], ...) {
  static int debug = GetPrivateProfileInt("debug", "log", 0, "mhuffyuv.ini");
  if (!debug) return;

  DWORD written;
  char buf[2000];
  va_list val;
  
  va_start(val, fmt);
  wvsprintf(buf, fmt, val);

  const COORD _80x50 = {80,50};
  static BOOL startup = (AllocConsole(), SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE), _80x50));
  WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), buf, lstrlen(buf), &written, 0);
}


int AppFlags() {
  static int flags = -1;
  if (flags < 0) {
    flags = 0;
    TCHAR apppath[MAX_PATH];
    if (GetModuleFileName(NULL, apppath, MAX_PATH)) {
      TCHAR* appname = strrchr(apppath, '\\');
      appname = appname ? appname+1 : apppath;
      Msg("App name is %s; ", appname);
      if (!lstrcmpi(appname, TEXT("premiere.exe")))
        flags = 1;
      if (!lstrcmpi(appname, TEXT("veditor.exe")))
        flags = 1;
      if (!lstrcmpi(appname, TEXT("afterfx.exe")))
        flags = 2;
      Msg("flags=%d\n", flags);
    }
  }
  return flags;
}


/********************************************************************
********************************************************************/

CodecInst* Open(ICOPEN* icinfo) {
  if (icinfo && icinfo->fccType != ICTYPE_VIDEO)
      return NULL;

  CodecInst* pinst = new CodecInst();

  if (icinfo) icinfo->dwError = pinst ? ICERR_OK : ICERR_MEMORY;

  return pinst;
}

DWORD Close(CodecInst* pinst) {
//    delete pinst;       // this caused problems when deleting at app close time
    return 1;
}

/********************************************************************
********************************************************************/

BOOL CodecInst::QueryAbout() { return TRUE; }

static BOOL CALLBACK AboutDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  if (uMsg == WM_COMMAND) {
    switch (LOWORD(wParam)) {
    case IDOK:
      EndDialog(hwndDlg, 0);
      break;
    case IDC_HOMEPAGE:
      ShellExecute(NULL, NULL, "http://www.amacias.tk", NULL, NULL, SW_SHOW);
      break;
    case IDC_EMAIL:
      ShellExecute(NULL, NULL, "mailto:amp_21004@yahoo.es", NULL, NULL, SW_SHOW);
      break;
    }
  }
  return FALSE;
}

DWORD CodecInst::About(HWND hwnd) {
  DialogBox(hmodulemhuffyuv, MAKEINTRESOURCE(IDD_ABOUT), hwnd, AboutDialogProc);
  return ICERR_OK;
}

static BOOL CALLBACK ConfigureDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  if (uMsg == WM_INITDIALOG) {
    CheckDlgButton(hwndDlg, IDC_COMPRESSASYUY2,
      GetPrivateProfileInt("general", "compress_as_yuy2", 0, "mhuffyuv.ini") ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton(hwndDlg, IDC_RGBOUTPUT,
      GetPrivateProfileInt("debug", "rgboutput", 0, "mhuffyuv.ini") ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton(hwndDlg, IDC_LOG,
      GetPrivateProfileInt("debug", "log", 0, "mhuffyuv.ini") ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton(hwndDlg, IDC_MOTION,
      GetPrivateProfileInt("debug", "motion", 0, "predictor.ini") ? BST_CHECKED : BST_UNCHECKED);
    CheckDlgButton(hwndDlg, IDC_ENSURE,
      GetPrivateProfileInt("debug", "ensure", 0, "predictor.ini") ? BST_CHECKED : BST_UNCHECKED);

  } else if (uMsg == WM_COMMAND) {
    switch (LOWORD(wParam)) {
    case IDOK:
      WritePrivateProfileString("general", "compress_as_yuy2",
        (IsDlgButtonChecked(hwndDlg, IDC_COMPRESSASYUY2) == BST_CHECKED) ? "1" : "0", "mhuffyuv.ini");
      WritePrivateProfileString("debug", "rgboutput",
        (IsDlgButtonChecked(hwndDlg, IDC_RGBOUTPUT) == BST_CHECKED) ? "1" : NULL, "mhuffyuv.ini");
      WritePrivateProfileString("debug", "log",
        (IsDlgButtonChecked(hwndDlg, IDC_LOG) == BST_CHECKED) ? "1" : "0", "mhuffyuv.ini");
      WritePrivateProfileString("debug", "motion",
        (IsDlgButtonChecked(hwndDlg, IDC_MOTION) == BST_CHECKED) ? "1" : "0", "predictor.ini");
      WritePrivateProfileString("debug", "ensure",
        (IsDlgButtonChecked(hwndDlg, IDC_ENSURE) == BST_CHECKED) ? "1" : "0", "predictor.ini");

    case IDCANCEL:
      EndDialog(hwndDlg, 0);
      break;
    default:
      return AboutDialogProc(hwndDlg, uMsg, wParam, lParam);    // handle email and home-page buttons
    }
  }
  return FALSE;
}

BOOL CodecInst::QueryConfigure() { return TRUE; }

DWORD CodecInst::Configure(HWND hwnd) {
  DialogBox(hmodulemhuffyuv, MAKEINTRESOURCE(IDD_CONFIGURE), hwnd, ConfigureDialogProc);
  return ICERR_OK;
}


// we have no state information which needs to be stored

DWORD CodecInst::GetState(LPVOID pv, DWORD dwSize) { return 0; }

DWORD CodecInst::SetState(LPVOID pv, DWORD dwSize) { return 0; }


DWORD CodecInst::GetInfo(ICINFO* icinfo, DWORD dwSize) {
  if (icinfo == NULL)
    return sizeof(ICINFO);

  if (dwSize < sizeof(ICINFO))
    return 0;

  icinfo->dwSize            = sizeof(ICINFO);
  icinfo->fccType           = ICTYPE_VIDEO;
  icinfo->fccHandler        = FOURCC_HFYU;
  icinfo->dwFlags           = 0;

  icinfo->dwVersion         = VERSION;
  icinfo->dwVersionICM      = ICVERSION;
  MultiByteToWideChar(CP_ACP, 0, szDescription, -1, icinfo->szDescription, sizeof(icinfo->szDescription)/sizeof(WCHAR));
  MultiByteToWideChar(CP_ACP, 0, szName, -1, icinfo->szName, sizeof(icinfo->szName)/sizeof(WCHAR));

  return sizeof(ICINFO);
}


/********************************************************************
********************************************************************/


struct PrintBitmapType {
  char s[32];
  PrintBitmapType(LPBITMAPINFOHEADER lpbi) {
    if (!lpbi)
      strcpy(s,  "(null)");
    else {
      *(DWORD*)s = lpbi->biCompression;
      s[4] = 0;
      if (!isalnum(s[0]) || !isalnum(s[1]) || !isalnum(s[2]) || !isalnum(s[3]))
        wsprintfA(s, "%x", lpbi->biCompression);
      wsprintfA(strchr(s, 0), ", bits = %d", lpbi->biBitCount);
    }
  }
};


// fast clipping of values to unsigned char range

static unsigned char clip[896];

static void InitClip() {
  memset(clip, 0, 320);
  for (int i=0; i<256; ++i) clip[i+320] = i;
  memset(clip+320+256, 255, 320);
}

/*
static inline unsigned char Clip(int x)
  { return clip[320 + ((x+0x8000) >> 16)]; }
*/

#define Clip(x)  clip[320 + (((x)+0x8000) >> 16)]


/********************************************************************
********************************************************************/


// 0=unknown, -1=compressed YUY2, -2=compressed RGB, -3=compressed RGBA, 1=YUY2, 2=UYVY, 3=RGB 24-bit, 4=RGB 32-bit
static int GetBitmapType(LPBITMAPINFOHEADER lpbi) {
  if (!lpbi)
    return 0;
  if (lpbi->biCompression == FOURCC_VYUY || lpbi->biCompression == FOURCC_YUY2)
    return 1;
  if (lpbi->biCompression == FOURCC_UYVY)
    return 2;
  if (lpbi->biCompression == 0 || lpbi->biCompression == ' BID')
    return (lpbi->biBitCount == 24) ? 3 : (lpbi->biBitCount == 32) ? 4 : 0;
  if (lpbi->biCompression == FOURCC_HFYU)
    return (lpbi->biBitCount == 16) ? -1 : (lpbi->biBitCount == 24) ? -2 : (lpbi->biBitCount == 32) ? -3 : 0;
  return 0;
}


static bool CanCompress(LPBITMAPINFOHEADER lpbiIn) {
  int intype = GetBitmapType(lpbiIn);
  return (intype == 1 || intype == 2 || intype == 3 || (intype == 4 && (AppFlags()&2)));
}


static bool CanCompress(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  if (!lpbiOut) return CanCompress(lpbiIn);
  int outtype = GetBitmapType(lpbiOut);
  switch (GetBitmapType(lpbiIn)) {
    case 1: case 2:
      return (outtype == -1);
    case 3:
      return (outtype == 1 || outtype == -1 || outtype == -2);
    case 4:
      return (outtype == -3 && (AppFlags() & 2));
    default:
      return false;
  }
}


/********************************************************************
********************************************************************/


DWORD CodecInst::CompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  Msg("CompressQuery: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));
  return CanCompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}


DWORD CodecInst::CompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  if (lpbiOut == 0)
    return sizeof(BITMAPINFOHEADER);

  if (!CanCompress(lpbiIn))
    return ICERR_BADFORMAT;

  bool compress_as_yuy2 = !!GetPrivateProfileInt("general", "compress_as_yuy2", false, "mhuffyuv.ini");

  *lpbiOut = *lpbiIn;
  lpbiOut->biCompression = FOURCC_HFYU;
  int intype = GetBitmapType(lpbiIn);
  lpbiOut->biBitCount = (intype==4) ? 32 : (intype==1 || intype==2 || compress_as_yuy2) ? 16 : 24;
  return ICERR_OK;
}


DWORD CodecInst::CompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
//  total_bpp = 0;
//  num_frames = 0;
  Msg("CompressBegin: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));

  CompressEnd();  // free resources if necessary

  enable_motion = !!GetPrivateProfileInt("debug", "motion", false, "predictor.ini");

  enable_ensure = !!GetPrivateProfileInt("debug", "ensure", false, "predictor.ini");
  if(enable_ensure)
	  enable_motion=1;


  yuy2_buffer[0] = new unsigned char[lpbiIn->biWidth * lpbiIn->biHeight * 2];
  yuy2_buffer[1] = new unsigned char[lpbiIn->biWidth * lpbiIn->biHeight * 2];
  icinfo_old = new unsigned char[lpbiIn->biWidth * lpbiIn->biHeight * 2];
  icinfo_sel=0;
  FrameCount=64;

  InitClip();

  return CanCompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}

DWORD CodecInst::CompressGetSize(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  // Assume 24 bpp worst-case for YUY2 input, 40 bpp for RGB.
  // The actual worst case is 43/51 bpp currently, but this is exceedingly improbable
  // (probably impossible with real captured video)
  return lpbiIn->biWidth * lpbiIn->biHeight * ((GetBitmapType(lpbiIn) <= 2) ? 3 : 5);
}




#define HUFF_ASM_FIRST(CHANNEL,INDEX,SPACING,INCREMENT) \
__asm  xor ebx,ebx     \
__asm  sub ebp,INCREMENT \
__asm  align 16        \
__asm loop_0:          \
__asm  mov bl,[esi+(INDEX+INCREMENT)] \
__asm  sub bl,[esi+(INDEX+INCREMENT-SPACING)] \
__asm  add esi,INCREMENT \
__asm  mov cl,CHANNEL##_shift[ebx] \
__asm  mov eax,CHANNEL##_add_shifted[ebx*4] \
__asm  xor ebx,ebx     \
__asm  add ch,cl       \
__asm  jl nostore_##INDEX \
                       \
__asm  sub cl,ch       \
__asm  add edi,4       \
__asm  shld edx,eax,cl \
__asm  add cl,ch       \
__asm  sub ch,32       \
__asm  mov [edi],edx   \
                       \
__asm nostore_##INDEX: \
__asm  shld edx,eax,cl

#define HUFF_ASM_NEXT(CHANNEL,INDEX,SPACING) \
__asm  mov bl,[esi+(INDEX)] \
__asm  sub bl,[esi+(INDEX-SPACING)] \
__asm  mov cl,CHANNEL##_shift[ebx] \
__asm  mov eax,CHANNEL##_add_shifted[ebx*4] \
__asm  xor ebx,ebx     \
__asm  add ch,cl       \
__asm  jl nostore_##INDEX \
                       \
__asm  sub cl,ch       \
__asm  add edi,4       \
__asm  shld edx,eax,cl \
__asm  add cl,ch       \
__asm  sub ch,32       \
__asm  mov [edi],edx   \
                       \
__asm nostore_##INDEX: \
__asm  shld edx,eax,cl

#define HUFF_ASM_END   \
__asm  cmp esi,ebp     \
__asm  jnz loop_0      \
                       \
__asm  cmp ch,-32      \
__asm  jle noextra     \
__asm  mov cl,ch       \
__asm  neg cl          \
__asm  add edi,4       \
__asm  shl edx,cl      \
__asm  mov [edi],edx   \
__asm noextra:


static inline int compress(unsigned char *input, DWORD* output, int len)
{
  unsigned char* end;
  DWORD *orgdest;
  orgdest = output;
  end = input + len;

__asm  mov esi,input
__asm  mov edi,output
__asm  mov ch,-32
__asm  push ebp   /* stupid compiler still uses EBP as frame pointer */
__asm  mov eax,[esi]
__asm  mov [edi],eax
__asm  mov ebp,end
__asm  align 16   /* this makes the important loop below be align 32 */

HUFF_ASM_FIRST(luma,0,2,4)
HUFF_ASM_NEXT(chroma,1,4)
HUFF_ASM_NEXT(luma,2,2)
HUFF_ASM_NEXT(chroma,3,4)
HUFF_ASM_END

__asm  pop ebp
__asm  mov output,edi

  return (DWORD(output) - DWORD(orgdest) + 4);
}//compress



DWORD CodecInst::Compress(ICCOMPRESS* icinfo, DWORD dwSize) 
{
  unsigned char* input = (unsigned char*)icinfo->lpInput;
  unsigned i,j,k,l,m,n,o,p;
  unsigned char *icinfo_ld;
  unsigned SizeImage=(icinfo->lpbiInput->biWidth * icinfo->lpbiInput->biHeight)<<1;

  if (GetBitmapType(icinfo->lpbiInput) == 3) 
  {
   ConvertRGB24toYUY2(input, yuy2_buffer[icinfo_sel], icinfo->lpbiInput->biWidth, icinfo->lpbiInput->biHeight);
   input=yuy2_buffer[icinfo_sel];
  }
  DWORD* output = (DWORD*)icinfo->lpOutput;

 if(enable_motion)
 {
  memcpy(icinfo_old,input,SizeImage);
  input=icinfo_old;

  icinfo_sel=(icinfo_sel + 1) & 1;
  
  unsigned EndLoop=SizeImage-8;

  if (GetBitmapType(icinfo->lpbiInput) == 3) 
   icinfo_ld=yuy2_buffer[icinfo_sel];
  else
   icinfo_ld=(unsigned char*)icinfo->lpPrev;
  if(icinfo_ld==NULL)
	  FrameCount=64;

      if((FrameCount==64) || (icinfo->dwFlags == ICCOMPRESS_KEYFRAME))
	  {
	   //insert a key frame every 64 frames
       *icinfo->lpdwFlags = AVIIF_KEYFRAME;
	   FrameCount=0;
	  }
	  else
	  {
       j=1;
	   k=2;
	   l=3;
	   m=4;
	   n=5;
	   o=6;
	   p=7;
	   // Substract the last frame from the actual
	   for(i=0; i<EndLoop;)
	   {
		  input[i] -= icinfo_ld[i];
		  i+=8;
		  input[j] -= icinfo_ld[j];
		  j+=8;
		  input[k] -= icinfo_ld[k];
		  k+=8;
		  input[l] -= icinfo_ld[l];
		  l+=8;
		  input[m] -= icinfo_ld[m];
		  m+=8;
		  input[n] -= icinfo_ld[n];
		  n+=8;
		  input[o] -= icinfo_ld[o];
		  o+=8;
		  input[p] -= icinfo_ld[p];
		  p+=8;
	   }//for
      }//else
      FrameCount++;
 }
 else
  *icinfo->lpdwFlags = AVIIF_KEYFRAME;
 

  icinfo->lpbiOutput->biSizeImage = compress(input,output,SizeImage);

  if(enable_ensure && !(*icinfo->lpdwFlags == AVIIF_KEYFRAME))
  {
	  unsigned Size;
      if (GetBitmapType(icinfo->lpbiInput) == 3) 
 	   Size=compress(yuy2_buffer[(icinfo_sel+1)&1],(DWORD*)input,SizeImage);
	  else
 	   Size=compress((unsigned char*)icinfo->lpInput,(DWORD*)input,SizeImage);

	  if(Size<icinfo->lpbiOutput->biSizeImage)
	  {
        icinfo->lpbiOutput->biSizeImage=Size;
		memcpy(output,input,Size);
        *icinfo->lpdwFlags = AVIIF_KEYFRAME;
		FrameCount=0;
	  }//if
  }//if



  if (icinfo->lpckid)
    *icinfo->lpckid = FOURCC_HFYU;



  return ICERR_OK;


}//Compress


inline void CodecInst::ConvertRGB24toYUY2(unsigned char* src, unsigned char* dst, int width, int height)
{
 unsigned char* rgb;
 unsigned char* yuv;
 int width3=width*3;
 int width2=width<<1;
 int y1,y2,scaled_y,b_y,r_y;
 


  for (int row = 0; row < height; ++row) 
  {
    rgb = src + width3 * (height-1-row);
    yuv = dst + width2 * row;
    for (int col = 0; col < width; col += 8) 
	{
      // y1 and y2 can't overflow
      yuv[0] = y1 = (cyb*rgb[0] + cyg*rgb[1] + cyr*rgb[2] + 0x108000) >> 16;
      yuv[2] = y2 = (cyb*rgb[3] + cyg*rgb[4] + cyr*rgb[5] + 0x108000) >> 16;
      scaled_y = (y1+y2 - 32) * int(255.0/219.0*32768+0.5);
      b_y = ((rgb[0]+rgb[3]) << 15) - scaled_y;
      r_y = ((rgb[2]+rgb[5]) << 15) - scaled_y;
      rgb += 6;
      yuv[1] = Clip((b_y >> 10) * int(1/2.018*1024+0.5) + 0x800000);  // u
      yuv[3] = Clip((r_y >> 10) * int(1/1.596*1024+0.5) + 0x800000);  // v
      yuv += 4;

      // y1 and y2 can't overflow
      yuv[0] = y1 = (cyb*rgb[0] + cyg*rgb[1] + cyr*rgb[2] + 0x108000) >> 16;
      yuv[2] = y2 = (cyb*rgb[3] + cyg*rgb[4] + cyr*rgb[5] + 0x108000) >> 16;
      scaled_y = (y1+y2 - 32) * int(255.0/219.0*32768+0.5);
      b_y = ((rgb[0]+rgb[3]) << 15) - scaled_y;
      r_y = ((rgb[2]+rgb[5]) << 15) - scaled_y;
      rgb += 6;
      yuv[1] = Clip((b_y >> 10) * int(1/2.018*1024+0.5) + 0x800000);  // u
      yuv[3] = Clip((r_y >> 10) * int(1/1.596*1024+0.5) + 0x800000);  // v
      yuv += 4;

      // y1 and y2 can't overflow
      yuv[0] = y1 = (cyb*rgb[0] + cyg*rgb[1] + cyr*rgb[2] + 0x108000) >> 16;
      yuv[2] = y2 = (cyb*rgb[3] + cyg*rgb[4] + cyr*rgb[5] + 0x108000) >> 16;
      scaled_y = (y1+y2 - 32) * int(255.0/219.0*32768+0.5);
      b_y = ((rgb[0]+rgb[3]) << 15) - scaled_y;
      r_y = ((rgb[2]+rgb[5]) << 15) - scaled_y;
      rgb += 6;
      yuv[1] = Clip((b_y >> 10) * int(1/2.018*1024+0.5) + 0x800000);  // u
      yuv[3] = Clip((r_y >> 10) * int(1/1.596*1024+0.5) + 0x800000);  // v
      yuv += 4;

      // y1 and y2 can't overflow
      yuv[0] = y1 = (cyb*rgb[0] + cyg*rgb[1] + cyr*rgb[2] + 0x108000) >> 16;
      yuv[2] = y2 = (cyb*rgb[3] + cyg*rgb[4] + cyr*rgb[5] + 0x108000) >> 16;
      scaled_y = (y1+y2 - 32) * int(255.0/219.0*32768+0.5);
      b_y = ((rgb[0]+rgb[3]) << 15) - scaled_y;
      r_y = ((rgb[2]+rgb[5]) << 15) - scaled_y;
      rgb += 6;
      yuv[1] = Clip((b_y >> 10) * int(1/2.018*1024+0.5) + 0x800000);  // u
      yuv[3] = Clip((r_y >> 10) * int(1/1.596*1024+0.5) + 0x800000);  // v
      yuv += 4;

    }
  }
}//ConvertRGB24toYUY2


DWORD CodecInst::CompressEnd() {


  if (yuy2_buffer[0]) 
  {
    delete[] yuy2_buffer[0];
    yuy2_buffer[0] = 0;
    delete[] yuy2_buffer[1];
    yuy2_buffer[1] = 0;
    delete[] icinfo_old;
    icinfo_old = 0;
  }
  icinfo_sel=3;

  return ICERR_OK;
}


/********************************************************************
********************************************************************/

static bool CanDecompress(LPBITMAPINFOHEADER lpbiIn) {
  int intype = GetBitmapType(lpbiIn);
  return (intype < 0 || intype == 1 || intype == 2);
}


static bool CanDecompress(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  if (!lpbiOut)
    return CanDecompress(lpbiIn);

  // must be 1:1 (no stretching)
  if (lpbiOut && (lpbiOut->biWidth != lpbiIn->biWidth || lpbiOut->biHeight != lpbiIn->biHeight))
    return false;

  int outtype = GetBitmapType(lpbiOut);
  switch (GetBitmapType(lpbiIn)) {
    case -1:
      // YUY2, RGB-24, RGB-32 output for compressed YUY2
      return (outtype == 1 || outtype == 3 || outtype == 4);
    case -2: case 1: case 2:
      // RGB-24, RGB-32 output only for YUY2/UYVY and compressed RGB
      return (outtype == 3 || outtype == 4);
    case -3:
      // RGB-32 output for compressed RGBA
      return (outtype == 4);
    default:
      return false;
  }
}

/********************************************************************
********************************************************************/


DWORD CodecInst::DecompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  Msg("DecompressQuery: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));
  return CanDecompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}


// This function should return "the output format which preserves the most
// information."  However, I now provide the option to return RGB format
// instead, since some programs treat the default format as the ONLY format.

DWORD CodecInst::DecompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {

  // if lpbiOut == NULL, then return the size required to hold an output format struct
  if (lpbiOut == NULL)
    return lpbiIn->biSize;

  if (!CanDecompress(lpbiIn))
    return ICERR_BADFORMAT;

  // get default rgboutput value and allow the user to override
  bool rgboutput = AppFlags() & 1;
  rgboutput = !!GetPrivateProfileInt("debug", "rgboutput", rgboutput, "mhuffyuv.ini");

  memcpy(lpbiOut, lpbiIn, lpbiIn->biSize);

  int intype = GetBitmapType(lpbiIn);
  if (intype == -3) {
    lpbiOut->biBitCount = 32;   // "RGBA" (though not really)
    lpbiOut->biCompression = 0;
    lpbiOut->biSizeImage = lpbiIn->biWidth * lpbiIn->biHeight * 4;
  } else if (rgboutput || intype == -2 || intype == 1 || intype == 2) {
    lpbiOut->biBitCount = 24;   // RGB
    lpbiOut->biCompression = 0;
    lpbiOut->biSizeImage = lpbiIn->biWidth * lpbiIn->biHeight * 3;
  } else {
    lpbiOut->biBitCount = 16;       // YUY2
    lpbiOut->biCompression = FOURCC_YUY2;
    lpbiOut->biSizeImage = lpbiIn->biWidth * lpbiIn->biHeight * 2;
  }

  return ICERR_OK;
}

DWORD CodecInst::DecompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  Msg("DecompressBegin: input = %s, output = %s\n", &PrintBitmapType(lpbiIn), &PrintBitmapType(lpbiOut));

  DecompressEnd();  // free resources if necessary

  // allocate buffer if decompressing HFYU->YUY2->RGB
  yuy2_buffer[0] = new unsigned char[lpbiIn->biWidth * lpbiIn->biHeight * 2];
  yuy2_buffer[1] = new unsigned char[lpbiIn->biWidth * lpbiIn->biHeight * 2];
  icinfo_sel=0;

  InitClip();

  return CanDecompress(lpbiIn, lpbiOut) ? ICERR_OK : ICERR_BADFORMAT;
}


// decode a single luma/chroma byte from the Huffman-coded bitstream
static inline unsigned char UnhuffLuma(const DWORD* in, unsigned& pos) {
  unsigned word = pos >> 5;
  unsigned bit = pos & 31;
  unsigned val = in[word];
  if (bit)    // the conditional is necessary because shifting by 32 doesn't work correctly on x86 processors
    val = (val << bit) | (in[word+1] >> (32-bit));
  // figure out the appropriate lookup table based on the number of leading zeros
  if (val >= (35 << 23)) {
    pos += decode_luma9[val >> 23].shift;
    return decode_luma9[val >> 23].add;
  }
  if (val >= (15 << 20)) {
    pos += decode_luma12[val >> 20].shift;
    return decode_luma12[val >> 20].add;
  }
  pos += decode_luma17[val >> 15].shift;
  return decode_luma17[val >> 15].add;
}

static inline unsigned char UnhuffChroma(const DWORD* in, unsigned& pos) {
  unsigned word = pos >> 5;
  unsigned bit = pos & 31;
  unsigned val = in[word];
  if (bit)
    val = (val << bit) | (in[word+1] >> (32-bit));
  if (val >= (2 << 23)) {
    pos += decode_chroma9[val >> 23].shift;
    return decode_chroma9[val >> 23].add;
  }
  if (val >= (4 << 16)) {
    pos += decode_chroma16[val >> 16].shift;
    return decode_chroma16[val >> 16].add;
  }
  if (val >= (26 << 9)) {
    pos += decode_chroma23[val >> 9].shift;
    return decode_chroma23[val >> 9].add;
  }
  pos += decode_chroma26[val >> 6].shift;
  return decode_chroma26[val >> 6].add;
}


static inline void YUV2RGB(int y, int u, int v, BYTE* out) {
  const int crv = int(1.596*65536+0.5);
  const int cgv = int(0.813*65536+0.5);
  const int cgu = int(0.391*65536+0.5);
  const int cbu = int(2.018*65536+0.5);

  int scaled_y = (y - 16) * int((255.0/219.0)*65536+0.5);

  out[0] = Clip(scaled_y + (u-128) * cbu); // blue
  out[1] = Clip(scaled_y - (u-128) * cgu - (v-128) * cgv); // green
  out[2] = Clip(scaled_y + (v-128) * crv); // red
}


// I originally tried to do this with a template function, but it produced
// incorrect code.  So, since you insist...

#define ConvertYUVtoRGB(uyvy,rgb_bytes) \
  const int width = lpbiOutput->biWidth; \
  const int height = lpbiOutput->biHeight; \
  lpbiOutput->biSizeImage = width * height * rgb_bytes; \
\
  unsigned char* const yuv_base = (unsigned char*)src; \
  unsigned char* const rgb_base = (unsigned char*)dst; \
  for (int row = 0; row < height; ++row) { \
    unsigned char* yuv = yuv_base + width * 2 * (height-1-row); \
    unsigned char* rgb = rgb_base + width * rgb_bytes * row; \
    YUV2RGB(yuv[0+uyvy], yuv[1-uyvy], yuv[3-uyvy], rgb); \
    YUV2RGB(yuv[2+uyvy], yuv[1-uyvy], yuv[3-uyvy], rgb+rgb_bytes); \
    if (rgb_bytes==4) \
      rgb[7] = rgb[3] = 0; \
    for (int col = 2; col < width; col += 2) { \
      yuv += 4; \
      rgb += rgb_bytes*2; \
      /* interpolate chroma values */ \
      YUV2RGB(yuv[0+uyvy], (int(yuv[1-uyvy])+int(yuv[-3-uyvy]))/2, (int(yuv[3-uyvy])+int(yuv[-1-uyvy]))/2, rgb); \
      YUV2RGB(yuv[2+uyvy], yuv[1-uyvy], yuv[3-uyvy], rgb+rgb_bytes); \
      if (rgb_bytes==4) \
        rgb[7] = rgb[3] = 0; \
    } \
  } \
  return ICERR_OK


static DWORD ConvertYUY2toRGB24(LPBITMAPINFOHEADER lpbiOutput, void* src, void* dst) { ConvertYUVtoRGB(false,3); }
static DWORD ConvertYUY2toRGB32(LPBITMAPINFOHEADER lpbiOutput, void* src, void* dst) { ConvertYUVtoRGB(false,4); }
static DWORD ConvertUYVYtoRGB24(LPBITMAPINFOHEADER lpbiOutput, void* src, void* dst) { ConvertYUVtoRGB(true,3); }
static DWORD ConvertUYVYtoRGB32(LPBITMAPINFOHEADER lpbiOutput, void* src, void* dst) { ConvertYUVtoRGB(true,4); }


DWORD CodecInst::Decompress(ICDECOMPRESS* icinfo, DWORD dwSize) 
{

  int width = icinfo->lpbiInput->biWidth;
  int height = icinfo->lpbiInput->biHeight;
  bool decompress_to_rgb = (icinfo->lpbiOutput->biBitCount > 16);

  const DWORD* const in = (DWORD*)icinfo->lpInput;
  unsigned char* out;
  out=yuy2_buffer[icinfo_sel];

  unsigned char* const end = out + width * height * 2;

 

  *(DWORD*)out = in[0];
  out += 4;
  unsigned pos = 32;
  while (out < end) {
    out[0] = out[-2] + UnhuffLuma(in, pos);
    out[1] = out[-3] + UnhuffChroma(in, pos);
    out += 2;
  }
      
	  out=yuy2_buffer[icinfo_sel];
	  icinfo_sel = (icinfo_sel + 1) & 1;
      unsigned SizeImage=end-out;
	  icinfo_old2=yuy2_buffer[icinfo_sel];

	  // Add the last frame to the actual if it is not a key frame
	  if(icinfo->dwFlags & ICDECOMPRESS_NOTKEYFRAME)
	  {
	   for(unsigned i=0; i<SizeImage; i++)
	 	  out[i] += icinfo_old2[i];
	  }

	  




  if (decompress_to_rgb) {
    if (icinfo->lpbiOutput->biBitCount == 32)
      ConvertYUY2toRGB32(icinfo->lpbiOutput, yuy2_buffer[icinfo_sel], icinfo->lpOutput);
    else
      ConvertYUY2toRGB24(icinfo->lpbiOutput, yuy2_buffer[icinfo_sel], icinfo->lpOutput);
  }

  return ICERR_OK;
}


DWORD CodecInst::DecompressRGB(ICDECOMPRESS* icinfo) {

  int width = icinfo->lpbiOutput->biWidth;
  int height = icinfo->lpbiOutput->biHeight;
  int size = width * height * (icinfo->lpbiOutput->biBitCount >> 3);
  icinfo->lpbiOutput->biSizeImage = size;

  const DWORD* const in = (DWORD*)icinfo->lpInput;
  unsigned char* out = (unsigned char*)icinfo->lpOutput;
  unsigned char* const end = out + size;

  *(DWORD*)out = in[0] >> 8;
  unsigned pos = 32;
  if (icinfo->lpbiInput->biBitCount == 32) {
    // RGBA
    out += 4;
    while (out < end) {
      out[0] = out[-4] + UnhuffLuma(in, pos);
      ++out;
    }
  } else if (icinfo->lpbiOutput->biBitCount == 24) {
    // RGB 24-bit
    out += 3;
    while (out < end) {
      out[0] = out[-3] + UnhuffLuma(in, pos);
      ++out;
    }
  } else {
    // RGB 32-bit (pad with zeroes)
    out += 4;
    while (out < end) {
      out[0] = out[-4] + UnhuffLuma(in, pos);
      out[1] = out[-3] + UnhuffLuma(in, pos);
      out[2] = out[-2] + UnhuffLuma(in, pos);
      out[3] = 0;
      out += 4;
    }
  }
  return ICERR_OK;
}


// palette-mapped output only
DWORD CodecInst::DecompressGetPalette(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut) {
  return ICERR_BADFORMAT;
}


DWORD CodecInst::DecompressEnd() {

// MModification
  if(icinfo_old2 != NULL)
  {
	  delete[] icinfo_old2;
	  icinfo_old2=NULL;
  }
// MModification

  if (yuy2_buffer[0]) 
  {
    delete[] yuy2_buffer[0];
    yuy2_buffer[0] = 0;
    delete[] yuy2_buffer[1];
    yuy2_buffer[1] = 0;
  }
  return ICERR_OK;
}

