//
// mhuffyuv v1.3.1, by Ben Rudiak-Gould.
// http://www.math.berkeley.edu/~benrg/mhuffyuv.html
//
// This file is copyright 2000 Ben Rudiak-Gould, and distributed under
// the terms of the GNU General Public License, v2 or later.  See
// http://www.gnu.org/copyleft/gpl.html.
//
// I edit these files in 10-point Verdana, a proportionally-spaced font.
// You may notice formatting oddities if you use a monospaced font.
//


#include <windows.h>
#include <vfw.h>
#pragma hdrstop


static const DWORD FOURCC_HFYU = mmioFOURCC('M','H','F','Y');   // our compressed format
static const DWORD FOURCC_YUY2 = mmioFOURCC('Y','U','Y','2');   // uncompressed YUY2
static const DWORD FOURCC_UYVY = mmioFOURCC('U','Y','V','Y');   // uncompressed UYVY
static const DWORD FOURCC_VYUY = mmioFOURCC('V','Y','U','Y');   // an alias for YUY2 used by ATI cards


extern HMODULE hmodulemhuffyuv;


// mhuffyuv.cpp

struct CodecInst {
  unsigned char *icinfo_old;
  unsigned char *icinfo_old2;
  int icinfo_sel;
  int FrameCount;
  unsigned char* yuy2_buffer[2];
  bool enable_motion;
  bool enable_ensure;

  // methods
  CodecInst();

  BOOL QueryAbout();
  DWORD About(HWND hwnd);

  BOOL QueryConfigure();
  DWORD Configure(HWND hwnd);

  DWORD GetState(LPVOID pv, DWORD dwSize);
  DWORD SetState(LPVOID pv, DWORD dwSize);

  DWORD GetInfo(ICINFO* icinfo, DWORD dwSize);

  DWORD CompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
  DWORD CompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
  DWORD CompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
  DWORD CompressGetSize(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
  DWORD Compress(ICCOMPRESS* icinfo, DWORD dwSize);
  DWORD CompressEnd();

  DWORD CompressYUY2(ICCOMPRESS* icinfo);
  DWORD CompressUYVY(ICCOMPRESS* icinfo);
  DWORD CompressRGB(ICCOMPRESS* icinfo);
  DWORD CompressRGBA(ICCOMPRESS* icinfo);

  void ConvertRGB24toYUY2(unsigned char* src, unsigned char* dst, int width, int height);

  DWORD DecompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
  DWORD DecompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
  DWORD DecompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
  DWORD Decompress(ICDECOMPRESS* icinfo, DWORD dwSize);
  DWORD DecompressGetPalette(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
  DWORD DecompressEnd();

  DWORD DecompressYUY2(ICDECOMPRESS* icinfo);
  DWORD DecompressRGB(ICDECOMPRESS* icinfo);

/*
  DWORD DrawQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
  DWORD DrawBegin(ICDRAWBEGIN* icinfo, DWORD dwSize);
  DWORD Draw(ICDRAW* icinfo, DWORD dwSize);
  DWORD DrawEnd();
  DWORD DrawWindow(PRECT prc);
*/
};

CodecInst* Open(ICOPEN* icinfo);
DWORD Close(CodecInst* pinst);


// tables.cpp

struct huffman_entry { unsigned char shift, add; };
extern const unsigned char luma_shift[256], chroma_shift[256];
extern const unsigned luma_add_shifted[256], chroma_add_shifted[256];
extern const huffman_entry decode_luma9[512], decode_luma12[512], decode_luma17[512];
extern const huffman_entry decode_chroma9[512], decode_chroma16[256];
extern const huffman_entry decode_chroma23[512], decode_chroma26[256];
